import struct, time, socket, binascii, pycom
from nodeLib import TempSensor
import config as cfg
import machine
from network import *

class Program:

    def __init__(self, firstRun):
    #-------------------------TEMPERATURE SENSOR---------------------------#
        
        self.temperatureSensor = TempSensor() 

    #-------------------------MODEL PARAMETERS-----------------------------#
        if (firstRun):
            self.expected_temp = cfg.device['expectedTemp']
            self.max_deviation = cfg.device['initialDeviation']
            self.deviceSleepTime = cfg.device['sleepTime']

            pycom.wifi_on_boot(False) #Turn off device wifi endpoint

    #-------------------------PRESERVE MODEL IN NVRAM AFTER DEEP SLEEP-----#    
        
            self.writeModelToNVRAM()

    #-------------------------PYCOM----------------------------------------#

        pycom.heartbeat(False) # disable the blue blinking
        pycom.rgbled(0x000000) #LED off
        

    #-------------------------INIT LORA------------------------------------#
    
    def initLora(self):

        self.lora = LoRa(mode=LoRa.LORAWAN)
        self.dev_eui = binascii.unhexlify(cfg.lora['dev_eui'])
        self.app_eui = binascii.unhexlify(cfg.lora['app_eui'])
        self.app_key = binascii.unhexlify(cfg.lora['app_key'])
        self.lora.add_channel(0, frequency=868100000, dr_min=0, dr_max=5)
        self.lora.add_channel(1, frequency=868100000, dr_min=0, dr_max=5)
        self.lora.add_channel(2, frequency=868100000, dr_min=0, dr_max=5)

        self.connectToLoRa()
        self.save_state()

    #-------------------------CONNECT TO LORA------------------------------#

    def connectToLoRa(self, blocking = False, timeout = 0, function = None):
        print("Connecting...")
        self.lora.join(activation=LoRa.OTAA, auth=(self.dev_eui, self.app_eui, self.app_key), timeout=0)

        if timeout == 0:
            while not self.lora.has_joined():
                if function == None:
                    time.sleep(2.5)
                else:
                    function()
        else:
            for x in range(timeout):
                if self.lora.has_joined():
                    break
                if function == None:
                    time.sleep(2.5)
                else:
                    function()
            
        if not self.lora.has_joined():
            return False

        for i in range(3, 16):
            self.lora.remove_channel(i)
            
        print("Connected!")
    
    #-------------------------REINIT AFTER DEEP SLEEP-----------------------#

    def reInit(self, restoredState):
        print("Reinitializing LoRA state from NVRAM")
        self.lora = restoredState

        #Set model from NVRAM
        self.max_deviation = pycom.nvs_get('dev') / 10
        self.expected_temp = pycom.nvs_get('temp')
        self.deviceSleepTime = pycom.nvs_get('sleep')

    #-------------------------MAKE SOCKET FOR SENDING-----------------------#

    def makeSocket(self, blocking):
        self.s = socket.socket(socket.AF_LORA, socket.SOCK_RAW)

        # set the LoRaWAN data rate
        self.s.setsockopt(socket.SOL_LORA, socket.SO_DR, 5)

        # make the socket blocking/non-blocking
        self.s.setblocking(blocking)

    #-------------------------SAVE STATE OF LORA BEFORE DEEP SLEEP----------#

    def save_state(self):
        print("Saving LoRA state to NVRAM")
        self.lora.nvram_save()
    
    
    def run(self):
        toSend = "A"*241
        self.sendTemp(toSend)
        self.tryToUpdateModel()
        self.save_state()
        self.deepSleep(self.deviceSleepTime)    
        # current_temp = self.temperatureSensor.readTemp()
        # if (not self.temperatureWithinAllowedRange(current_temp)):
        #     self.sendTemp(current_temp)
        #     self.tryToUpdateModel()
        #     self.save_state()
        #     self.deepSleep(self.deviceSleepTime)
        # else:
        #     print("CurrentTemp: {0}. PreviousTemp: {1}. MaxDeviation: {2}. Data not sent".format(current_temp, self.expected_temp, self.max_deviation))
        #     self.deepSleep(self.deviceSleepTime)

    
    def temperatureWithinAllowedRange(self,current_temp):
        return (abs(float(current_temp) - float(self.expected_temp)) < self.max_deviation)

    def sendTemp(self,current_temp):
        print("Sending measurement: {0} of {1} bytes".format(current_temp, len(current_temp)))
        self.makeSocket(True)
        self.s.send(current_temp)

    def recvFromServer(self):
        print("Trying to receive packet from server...")
        self.makeSocket(False)
        rx, port = self.s.recvfrom(256)
        return rx, port

    def tryToUpdateModel(self):
        rx, port = self.recvFromServer()
        if rx:
            print("Received {0} from server on port {1}".format(rx,port))
            model = rx.decode('utf-8').split(",")
            if (len(model) > 0):
                self.updateModel(float(model[0]), int(model[1]), float(model[2]))
        else:
            print("Nothing received from the server")

    def updateModel(self, deviation, sleeptime, expectedTemperature):
        self.max_deviation = deviation
        self.deviceSleepTime = sleeptime
        self.expected_temp = expectedTemperature
        self.writeModelToNVRAM()
        print("Model updated. MaxDeviation: {0}. Device sleep time: {1}. Expected temperature {2}.".format(self.max_deviation,self.deviceSleepTime,self.expected_temp))

    def deepSleep(self,secondsToSleep):
        print("Going into deepsleep for {0} seconds".format(secondsToSleep))
        self.save_state()
        machine.deepsleep(secondsToSleep*1000)

    def writeModelToNVRAM(self):
        #only 32-bit integers are allowed
        max_dev = int(self.max_deviation * 10)
        temp = round(int(self.expected_temp))
        
        pycom.nvs_set('dev',max_dev)
        pycom.nvs_set('temp',temp)
        pycom.nvs_set('sleep',self.deviceSleepTime)